"""
    This file is part of Accountant.

    Accountant is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Accountant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with Accountant.  If not, see <http://www.gnu.org/licenses/>.
"""
from datetime import date

from sqlalchemy import func, case, cast
from sqlalchemy.orm import validates

from accountant import db

from .entries import Entry
from .operations import Operation


class Account(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(200), nullable=False)
    authorized_overdraft = db.Column(db.Integer, nullable=True, default=0)

    def __init__(self, name, authorized_overdraft):
        self.name = name
        self.authorized_overdraft = authorized_overdraft

    @classmethod
    def get_accounts(cls, session):
        query = session.query(
            cls.id.label("id"),
            cls.name.label("name"),
            cls.authorized_overdraft.label("authorized_overdraft"),
            func.sum(Entry.value).label("future"),
            func.sum(case([(Entry.pointed, Entry.value,)],
                          else_=cast(0, db.Numeric(15, 2)))).label("pointed"),
            func.sum(case([(Entry.operation_date < func.now(), Entry.value,)],
                          else_=cast(0, db.Numeric(15, 2)))).label("current")
        ).outerjoin(Entry).group_by(cls.id).order_by(cls.id)

        return query

    @classmethod
    def get(cls, session, account, begin=date.today(), end=None):
        if isinstance(account, int) or isinstance(account, str):
            account_id = account
        else:
            account_id = account.id

        current = begin if end is None else end

        query = session.query(
            cls,
            func.sum(Operation.value).label("future"),
            func.sum(case([(Operation.pointed,
                            Operation.value,)],
                          else_=0)).label("pointed"),
            func.sum(case([(Operation.operation_date <= str(current),
                            Operation.value,)],
                          else_=0)).label("current")
        ).group_by(cls.id)

        if end:
            subquery = session.query(
                Operation.account_id,
                func.sum(case([(func.sign(Operation.value) == -1,
                                Operation.value)],
                              else_=0)).label("expenses"),
                func.sum(case([(func.sign(Operation.value) == 1,
                                Operation.value)],
                              else_=0)).label("revenues"),
                func.sum(Operation.value).label("balance")
            ).filter(
                Operation.operation_date >= begin,
                Operation.operation_date <= end
            ).group_by(
                Operation.account_id
            ).subquery()

            base_subquery = query.subquery()

            query = session.query(
                base_subquery, subquery
            ).select_from(base_subquery, subquery).filter(
                base_subquery.c.id == subquery.c.account_id,
                base_subquery.c.id == account_id,
            )
        else:
            query = query.filter(cls.id == account_id)

        return query.one()

    @validates('authorized_overdraft')
    def validate_authorized_overdraft(self, key, authorized_overdraft):
        assert authorized_overdraft > 0

        return authorized_overdraft
